/*
 * @(#)PlainDocument.java	1.34 00/02/02
 *
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package ginian.javaEditorPane;

import java.awt.*;
import javax.swing.*;
import javax.swing.text.*;
import java.util.*;
import javax.swing.event.*;
import java.awt.font.TextAttribute;

public class JavaDocument extends DefaultStyledDocument {

	/**
	 * Normal text token id. This should be used to mark
	 * normal text.
	 */
	public static final String REGULAR = "0";

	/**
	 * Comment 1 token id. This can be used to mark a comment.
	 */
	public static final String COMMENT1 = "1";

	/**
	 * Comment 2 token id. This can be used to mark a comment.
	 */
	public static final String COMMENT2 = "2";
	/**
	 * Literal 1 token id. This can be used to mark a string
	 * literal (eg, C mode uses this to mark "..." literals)
	 */
	public static final String LITERAL1 = "3";

	/**
	 * Literal 2 token id. This can be used to mark an object
	 * literal (eg, Java mode uses this to mark true, false, etc)
	 */
	public static final String LITERAL2 = "4";

	/**
	 * Label token id. This can be used to mark labels
	 * (eg, C mode uses this to mark ...: sequences)
	 */
	public static final String LABEL = "5";

	/**
	 * Keyword 1 token id. This can be used to mark a
	 * keyword. This should be used for general language
	 * constructs.
	 */
	public static final String KEYWORD1 = "6";

	/**
	 * Keyword 2 token id. This can be used to mark a
	 * keyword. This should be used for preprocessor
	 * commands, or variables.
	 */
	public static final String KEYWORD2 = "7";

	/**
	 * Keyword 3 token id. This can be used to mark a
	 * keyword. This should be used for data types.
	 */
	public static final String KEYWORD3 = "8";

	/**
	 * Operator token id. This can be used to mark an
	 * operator. (eg, SQL mode marks +, -, etc with this
	 * token type)
	 */
	public static final String OPERATOR = "9";

	/**
	 * Invalid token id. This can be used to mark invalid
	 * or incomplete tokens, so the user can easily spot
	 * syntax errors.
	 */
	public static final String INVALID = "10";
	public static final String SPACE = "11";
  private HashMap _keywords;

  public JavaDocument() {
    super();
    initStyles();
    _keywords = createKeywords();
  }
  
  public HashMap createKeywords() {
    HashMap javaKeywords = new HashMap();
    javaKeywords.put("package",KEYWORD2);
    javaKeywords.put("import",KEYWORD2);
    javaKeywords.put("byte",KEYWORD3);
    javaKeywords.put("char",KEYWORD3);
    javaKeywords.put("short",KEYWORD3);
    javaKeywords.put("int",KEYWORD3);
    javaKeywords.put("long",KEYWORD3);
    javaKeywords.put("float",KEYWORD3);
    javaKeywords.put("double",KEYWORD3);
    javaKeywords.put("boolean",KEYWORD3);
    javaKeywords.put("void",KEYWORD3);
    javaKeywords.put("class",KEYWORD3);
    javaKeywords.put("interface",KEYWORD3);
    javaKeywords.put("abstract",KEYWORD1);
    javaKeywords.put("final",KEYWORD1);
    javaKeywords.put("private",KEYWORD1);
    javaKeywords.put("protected",KEYWORD1);
    javaKeywords.put("public",KEYWORD1);
    javaKeywords.put("static",KEYWORD1);
    javaKeywords.put("synchronized",KEYWORD1);
    javaKeywords.put("volatile",KEYWORD1);
    javaKeywords.put("transient",KEYWORD1);
    javaKeywords.put("break",KEYWORD1);
    javaKeywords.put("case",KEYWORD1);
    javaKeywords.put("continue",KEYWORD1);
    javaKeywords.put("default",KEYWORD1);
    javaKeywords.put("do",KEYWORD1);
    javaKeywords.put("else",KEYWORD1);
    javaKeywords.put("for",KEYWORD1);
    javaKeywords.put("if",KEYWORD1);
    javaKeywords.put("instanceof",KEYWORD1);
    javaKeywords.put("new",KEYWORD1);
    javaKeywords.put("return",KEYWORD1);
    javaKeywords.put("switch",KEYWORD1);
    javaKeywords.put("while",KEYWORD1);
    javaKeywords.put("throw",KEYWORD1);
    javaKeywords.put("try",KEYWORD1);
    javaKeywords.put("catch",KEYWORD1);
    javaKeywords.put("extends",KEYWORD1);
    javaKeywords.put("finally",KEYWORD1);
    javaKeywords.put("implements",KEYWORD1);
    javaKeywords.put("throws",KEYWORD1);
    javaKeywords.put("this",LITERAL2);
    javaKeywords.put("null",LITERAL2);
    javaKeywords.put("super",LITERAL2);
    javaKeywords.put("true",LITERAL2);
    javaKeywords.put("false",LITERAL2);

    return javaKeywords;
  }

  protected  void initStyles() {
    Style def = StyleContext.getDefaultStyleContext().
    getStyle(StyleContext.DEFAULT_STYLE);

    Style regular = addStyle(REGULAR, def);
    StyleConstants.setFontFamily(def, "monospaced");
    StyleConstants.setFontSize(regular,12); 
    
    Style space = addStyle(SPACE,regular);
    
    Style invalid = addStyle(INVALID, regular);
    StyleConstants.setBold(invalid, true);
    StyleConstants.setForeground(invalid, Color.black);
    StyleConstants.setBackground(invalid,Color.red);
    
    Style keyword1 = addStyle(KEYWORD1, regular);
    StyleConstants.setBold(keyword1, true);
    StyleConstants.setForeground(keyword1, Color.black);
 
    Style keyword2 = addStyle(KEYWORD2, regular);
    StyleConstants.setBold(keyword2, true);
    StyleConstants.setForeground(keyword2, Color.magenta);
    
    Style keyword3 = addStyle(KEYWORD3, regular);
    StyleConstants.setBold(keyword3, true);
    StyleConstants.setForeground(keyword3, Color.blue);   
    
    Style literal2 = addStyle(LITERAL2, regular);
    StyleConstants.setBold(literal2, true);
    StyleConstants.setForeground(literal2, new Color(150,0,0));    

    Style comment1 = addStyle(COMMENT1, regular);
    StyleConstants.setItalic(comment1, true);
    StyleConstants.setForeground(comment1, new Color(0,150,0));   
    
    Style literal1 = addStyle(LITERAL1, regular);
    StyleConstants.setBold(literal1, true);
    StyleConstants.setForeground(literal1, Color.red);   
  }
  
public void insertString(int offset, String str, AttributeSet a) throws BadLocationException {
  super.insertString(offset, str,getStyle(REGULAR));
  scanRange(offset,str.length());
}
public void remove(int offs,int len) throws BadLocationException {
  super.remove(offs,len);
  scanRange(offs,len);
}
protected void scanRange(int offset,int len) throws BadLocationException {
  Element root = getDefaultRootElement();
  int firstParagraph = root.getElementIndex(offset);
  int lastParagraph = root.getElementIndex(offset+len);

  for(int i=0;i<1+lastParagraph-firstParagraph;i++)
    scanParagraph(root.getElement(firstParagraph+i));
}
protected void scanParagraph(Element paragraph) throws BadLocationException {
  int startPosition = paragraph.getStartOffset();
  int length = (paragraph.getEndOffset()<getLength() ? paragraph.getEndOffset() : getLength())-startPosition;
  String content = getText(startPosition,length);

  for(int wordIndex=0;wordIndex<content.length();)  {
    char indexedChar = content.charAt(wordIndex);
  /*if(Character.isWhitespace(indexedChar)) {
      setCharacterAttributes(startPosition+wordIndex,1,getStyle(SPACE),true);
      wordIndex++;    
    }
    else*/ 
    if(Character.isJavaIdentifierStart(indexedChar)) {
      String scannedIdentifier = scanIdentifier(content,wordIndex);
      String tokenType = (String)_keywords.get(scannedIdentifier);
      int scannedIdentifierLength = scannedIdentifier.length();
      AttributeSet style;
      if(tokenType==null)
        style = getStyle(REGULAR);
      else
        style = getStyle(tokenType);
      setCharacterAttributes(startPosition+wordIndex,scannedIdentifierLength,style,true);
      wordIndex=wordIndex+scannedIdentifierLength;     
    }
    else if(isStringLiteralStart(indexedChar)) {
      int scannedStringLength = scanStringLiteral(content,wordIndex);
      setCharacterAttributes(startPosition+wordIndex,scannedStringLength,getStyle(LITERAL1),true);
      wordIndex=wordIndex+scannedStringLength;      
    }
    else if(isCharLiteralStart(indexedChar)) {
      int scannedCharLength = scanCharLiteral(content,wordIndex);
      setCharacterAttributes(startPosition+wordIndex,scannedCharLength,getStyle(LITERAL1),true);
      wordIndex=wordIndex+scannedCharLength;      
    }    
    else if(isSingleLineCommentStart(content,wordIndex)) {
      int scannedCommentLength = scanSingleLineComment(content,wordIndex);
      setCharacterAttributes(startPosition+wordIndex,scannedCommentLength,getStyle(COMMENT1),true);
      wordIndex=wordIndex+scannedCommentLength;       
    }
    else {
      setCharacterAttributes(startPosition+wordIndex,1,getStyle(REGULAR),true);
      wordIndex++;
    }
  }
  }
  protected boolean isSingleLineCommentStart(String string,int index) {
    return string.regionMatches(index,"//",0,2);
  }
  protected int scanSingleLineComment(String string,int index) {
      int commentLength = 0;
      for(commentLength = 0;commentLength<(string.length()-index);commentLength++) {
        char commentChar = string.charAt(index+commentLength);
        if(commentChar=='\n')
          break;
      }
      return commentLength;
  }
  protected boolean isStringLiteralStart(char c) {
    return (c=='"');
  }
  protected int scanStringLiteral(String string,int index) {
      int stringLength = 1;
      boolean backslash = false;
      for(stringLength = 1;stringLength<(string.length()-index);stringLength++) {
        char stringChar = string.charAt(index+stringLength);
        if(stringChar=='\\')
          backslash=true;
        else if(stringChar=='\n')
          break;
        else if(stringChar=='"' && !backslash) {
          stringLength++;
          break;          
        }
        else if(backslash)
          backslash=false;
      }
      return stringLength;
  }
  protected boolean isCharLiteralStart(char c) {
    return (c=='\'');
  }  
  protected int scanCharLiteral(String string,int index) {
      int charLength = 1;
      boolean backslash = false;
      for(charLength = 1;charLength<(string.length()-index);charLength++) {
        char charChar = string.charAt(index+charLength);
        if(charChar=='\\')
          backslash=true;
        else if(charChar=='\n')
          break;
        else if(charChar=='\'' && !backslash) {
          charLength++;
          break;          
        }
        else if(backslash)
          backslash=false;
      }
      return charLength;
  }
  protected String scanIdentifier(String string,int index) {
    int identifierLength = 0;
    if(!Character.isJavaIdentifierStart(string.charAt(index)))
      return "";
    else
      for(identifierLength = 1;identifierLength<(string.length()-index);identifierLength++) {
        char identifierChar = string.charAt(index+identifierLength);
        if(!Character.isJavaIdentifierPart(string.charAt(index+identifierLength)))
          break;
      }
      return string.substring(index,index+identifierLength);
  }
}